import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class CheckSumRandom {
  // the number of producer and consumer threads
  private static final int THREADS = 1000;
  // the number of items each producer adds and each consumer gets from the queue
  private static final int EACH = 1000000;
  // the capacity of the queue
  private static final int CAPACITY = 20;
  public static void main(String[] argv) {
    var queue = new ProducerConsumerQueue<Integer>(CAPACITY);
    var random = new Random();
    var producers = new ArrayList<TestProducer>();
    var consumers = new ArrayList<TestConsumer>();
    var currentlyRunning = new AtomicInteger(0);
    var threadPool = Executors.newCachedThreadPool();
    var syncer = new StartupSyncer(THREADS * 2);
    for (int i = 0; i < THREADS; i++) {
      var producer = new TestProducer(EACH, queue, currentlyRunning, syncer, random);
      threadPool.submit(producer);
      producers.add(producer);
      var consumer = new TestConsumer(EACH, queue, currentlyRunning, syncer);
      threadPool.submit(consumer);
      consumers.add(consumer);
    }
    try {
      threadPool.shutdown();
      threadPool.awaitTermination(1000, TimeUnit.SECONDS);
      if (!threadPool.isTerminated()) {
        System.out.println("Test failed: thread still running");
        System.exit(-1);
      }
    } catch (InterruptedException ignored) {}
    // if they all finished, check:
    var produced = producers.stream().map(TestProducer::getCheckSum).reduce(Long::sum);
    var consumed = consumers.stream().map(TestConsumer::getCheckSum).reduce(Long::sum);
    System.out.println(produced.equals(consumed));

  }

  public static class TestProducer implements Runnable {
    private final int toProduce;
    private final ProducerConsumerQueue<Integer> queue;
    private final AtomicInteger currentlyRunning;
    private final StartupSyncer syncer;
    private final Random random;
    private long checkSum = 0;

    public TestProducer(int toProduce, ProducerConsumerQueue<Integer> queue, AtomicInteger currentlyRunning, StartupSyncer syncer, Random random) {
      this.toProduce = toProduce;
      this.queue = queue;
      this.currentlyRunning = currentlyRunning;
      this.syncer = syncer;
      this.random = random;
    }

    @Override
    public void run() {
      currentlyRunning.incrementAndGet();
      try {
        syncer.await();
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      }
      try {
        for (int i = 0; i < toProduce; i++) {
          var next = random.nextInt();
          queue.add(next);
          checkSum += next;
          Thread.yield();
        }
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      } finally {
        currentlyRunning.decrementAndGet();
      }
    }

    public long getCheckSum() {
      return checkSum;
    }
  }

  public static class TestConsumer implements Runnable {
    private final int toConsume;
    private final ProducerConsumerQueue<Integer> queue;

    private final AtomicInteger currentlyRunning;
    private final StartupSyncer syncer;
    private long checkSum = 0;

    public TestConsumer(int toConsume, ProducerConsumerQueue<Integer> queue, AtomicInteger currentlyRunning, StartupSyncer syncer) {
      this.toConsume = toConsume;
      this.queue = queue;
      this.currentlyRunning = currentlyRunning;
      this.syncer = syncer;
    }

    @Override
    public void run() {
      currentlyRunning.incrementAndGet();
      try {
        syncer.await();
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      }
      var atStart = currentlyRunning.get();
      System.out.println("Currently running: " + atStart);
      try {
        for (int i = 0; i < toConsume; i++) {
          var next = queue.get();
          checkSum += next;
          Thread.yield();
        }
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      } finally {
        currentlyRunning.decrementAndGet();
      }
    }

    public long getCheckSum() {
      return checkSum;
    }
  }
}
