import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class OrderingMetadata {
  // the number of producer and consumer threads
  private static final int THREADS = 1000;
  // the number of items each producer adds and each consumer gets from the queue
  private static final int EACH = 1000;
  // the capacity of the queue
  private static final int CAPACITY = 20;
  public static void main(String[] argv) {
    var queue = new ProducerConsumerQueue<OrderingTestData<Integer>>(CAPACITY);
    var producers = new ArrayList<TestProducer>();
    var consumers = new ArrayList<TestConsumer>();
    var currentlyRunning = new AtomicInteger(0);
    var threadPool = Executors.newCachedThreadPool();
    var syncer = new StartupSyncer(THREADS * 2);
    for (int i = 0; i < THREADS; i++) {
      var producer = new TestProducer(EACH, i, queue, currentlyRunning, syncer);
      threadPool.submit(producer);
      producers.add(producer);
      var consumer = new TestConsumer(EACH, queue, currentlyRunning, syncer);
      threadPool.submit(consumer);
      consumers.add(consumer);
    }
    try {
      threadPool.shutdown();
      threadPool.awaitTermination(1000, TimeUnit.SECONDS);
      if (!threadPool.isTerminated()) {
        System.out.println("Test failed: thread still running");
        System.exit(-1);
      }
    } catch (InterruptedException ignored) {}
    // if they all finished, check:
    var produced = producers.stream().map(TestProducer::getCheckSum).reduce(Long::sum);
    var consumed = consumers.stream().map(TestConsumer::getCheckSum).reduce(Long::sum);
    System.out.println(produced.equals(consumed));

  }

  public static class TestProducer implements Runnable {
    private final int toProduce;
    private final ProducerConsumerQueue<OrderingTestData<Integer>> queue;
    private final AtomicInteger currentlyRunning;
    private final StartupSyncer syncer;
    private long checkSum = 0;
    private int id;

    public TestProducer(int toProduce, int id, ProducerConsumerQueue<OrderingTestData<Integer>> queue, AtomicInteger currentlyRunning, StartupSyncer syncer) {
      this.toProduce = toProduce;
      this.id = id;
      this.queue = queue;
      this.currentlyRunning = currentlyRunning;
      this.syncer = syncer;
    }

    @Override
    public void run() {
      var random = ThreadLocalRandom.current();
      currentlyRunning.incrementAndGet();
      try {
        syncer.await();
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      }
      try {
        for (int i = 0; i < toProduce; i++) {
          var next = random.nextInt();
          var data = new OrderingTestData<>(next, id, i);
          queue.add(data);
          checkSum += next;
        }
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      } finally {
        currentlyRunning.decrementAndGet();
      }
    }

    public long getCheckSum() {
      return checkSum;
    }
  }

  public static class TestConsumer implements Runnable {
    private final int toConsume;
    private final ProducerConsumerQueue<OrderingTestData<Integer>> queue;

    private final AtomicInteger currentlyRunning;
    private final StartupSyncer syncer;
    private long checkSum = 0;

    public TestConsumer(int toConsume, ProducerConsumerQueue<OrderingTestData<Integer>> queue, AtomicInteger currentlyRunning, StartupSyncer syncer) {
      this.toConsume = toConsume;
      this.queue = queue;
      this.currentlyRunning = currentlyRunning;
      this.syncer = syncer;
    }

    @Override
    public void run() {
      currentlyRunning.incrementAndGet();
      try {
        syncer.await();
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      }
      try {
        Map<Integer,Integer> lastSeen = new HashMap<>();
        for (int i = 0; i < toConsume; i++) {
          var next = queue.get();
          var prevSeen = lastSeen.getOrDefault(next.getProducer(), Integer.MIN_VALUE);
          if (prevSeen >= next.getTimestamp()) {
            System.err.println("Out of order from producer " + next.getProducer());
            throw new RuntimeException("Out of order from producer " + next.getProducer());
          }
          checkSum += next.getValue();
          lastSeen.put(next.getProducer(), next.getTimestamp());
        }
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      } finally {
        currentlyRunning.decrementAndGet();
      }
    }

    public long getCheckSum() {
      return checkSum;
    }
  }
}
