import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class PCQWithMetaData<T> {
  private final TimeStamped<T>[] items;
  private int nextInsert = 0;
  private int nextRemove = 0;
  private int count = 0;
  private final Lock lock = new ReentrantLock();
  private final Condition notEmpty = lock.newCondition();
  private final Condition notFull = lock.newCondition();


  PCQWithMetaData(int capacity) {
    items = (TimeStamped<T>[]) new TimeStamped[capacity];
  }

  // if the queue is full, the calling thread should wait until there is space in the queue
  public void add(T element) throws InterruptedException {
    try {
      lock.lock();
      while (isFull()) {
        notFull.await();
      }
      items[nextInsert] = new TimeStamped<>(element);
      nextInsert = (nextInsert + 1) % items.length;
      count++;
      notEmpty.signal();
    } finally {
      lock.unlock();
    }
  }

  // if the queue is empty, the calling thread should wait until there is an item in the queue
  public T get() throws InterruptedException {
    try {
      lock.lock();
      while (isEmpty()) {
        notEmpty.await();
      }
      var element = items[nextRemove];
      nextRemove = (nextRemove + 1) % items.length;
      count--;
      notFull.signal();
      return element.value;
    } finally {
      lock.unlock();
    }
  }

  private boolean isFull() {
    return count == items.length;
  }

  private boolean isEmpty() {
    return count == 0;
  }

  private static class TimeStamped<U> {
    U value;
    long timestamp;

    public TimeStamped(U value) {
      this.value = value;
      this.timestamp = System.nanoTime();
    }
  }

  public boolean checkTimestamps() {
    try {
      lock.lock();
      long lastTimeStamp = Long.MIN_VALUE;
      for (int i = 0; i < count; i++) {
        int idx = (nextRemove + i) % items.length;
        if (items[idx].timestamp < lastTimeStamp) {
          return false;
        }
        lastTimeStamp = items[idx].timestamp;
      }
    } finally {
      lock.unlock();
    }
    return true;
  }


}
